from typing import Tuple
import torch
import torch_scatter
import torch.multiprocessing as mp
from abc import ABC

class BaseSampler(ABC):
    r"""An abstract base class that initializes a graph sampler and provides
    :meth:`sample_from_nodes` and :meth:`sample_from_edges` routines.
    """
    def sample_from_nodes(
        self,
        node:int, 
        edge_index:torch.Tensor, 
        num_nodes:int, 
        **kwargs
    ) -> Tuple[torch.tensor, torch.tensor]:
        r"""Performs sampling from the nodes specified in: node,
        returning a sampled subgraph in the specified output format: Tuple[int, torch.tensor].

        Args:
            node: the seed node index
            edge_index: edges in the graph
            num_nodes: the num of all 
            **kwargs: other kwargs
        Returns:
            samples_nodes: the node sampled
            edge_index: the edge sampled
        """
        raise NotImplementedError
    
    def sample_from_nodes(
        self,
        nodes:torch.Tensor, 
        edge_index:torch.Tensor, 
        num_nodes:int, 
        **kwargs
    ) -> Tuple[torch.Tensor, torch.tensor]:
        r"""Performs sampling from the nodes specified in: node,
        returning a sampled subgraph in the specified output format: Tuple[int, torch.tensor].

        Args:
            nodes: the seed nodes index
            edge_index: edges in the graph
            num_nodes: the num of all 
            **kwargs: other kwargs
        Returns:
            samples_nodes: the node sampled
            edge_index: the edge sampled
        """
        raise NotImplementedError

class NeighborSampler(BaseSampler):
    def __init__(self) -> None:
        super().__init__()

    def sample_from_node(
        self, 
        node: int, 
        edge_index: torch.Tensor, 
        num_nodes: int, 
        fanout: int
    ) -> Tuple[torch.Tensor, torch.tensor]:
        r"""Performs sampling from the nodes specified in: node,
        returning a sampled subgraph in the specified output format: Tuple[int, torch.tensor].

        Args:
            node: the seed node index
            edge_index: edges in the graph
            num_nodes: the num of all node
            fanout: the number of max neighbor chosen
        Returns:
            samples_nodes: the node sampled
            edge_index: the edge sampled
        """
        row, col = edge_index
        deg = torch_scatter.scatter_add(torch.ones_like(row), row, dim=0, dim_size=num_nodes)
        neighbors=torch.stack([row[row==node],col[row==node]],dim=0)
        print('neighbors: \n', neighbors)
        if deg[node]<=fanout:
            return torch.unique(neighbors[1], dim=0), neighbors
        else:
            random_index = torch.multinomial(torch.ones(deg[node]), fanout, replacement=False)# torch.randperm(neighbors.shape[1])[0:fanout]
            print("random_index:\n", random_index)
            edge_index = neighbors.index_select(dim = 1, index=random_index)
            samples_nodes = torch.unique(edge_index.view(-1), dim=0)
            return samples_nodes, edge_index

    def sample_from_nodes(
        self, 
        nodes: torch.Tensor, 
        edge_index: torch.Tensor, 
        num_nodes: int, 
        fanout: int
    ) -> Tuple[torch.Tensor, torch.tensor]:
        r"""Performs sampling from the nodes specified in: node,
        returning a sampled subgraph in the specified output format: Tuple[int, torch.tensor].

        Args:
            node: the seed node index
            edge_index: edges in the graph
            num_nodes: the num of all node
            fanout: the number of max neighbor chosen
        Returns:
            nodes: the node sampled
            edge_index: the edge sampled
        """
        samples_nodes=torch.IntTensor([])
        row=torch.IntTensor([])
        col=torch.IntTensor([])
        with mp.Pool(processes=torch.get_num_threads()) as p:
            results = [p.apply_async(self.sample_from_node, 
                                     (node, edge_index, num_nodes, fanout)) 
                       for node in nodes]
            for result in results:
                samples_nodes_i, edge_index_i = result.get()
                samples_nodes = torch.unique(torch.cat([samples_nodes, samples_nodes_i]))
                row = torch.cat([row, edge_index_i[0]])
                col = torch.cat([col, edge_index_i[1]])
        # 单线程循环法：
        # for node in nodes:
        #     samples_nodes_i,edge_index_i = self.sample_from_node(node, edge_index, num_nodes, fanout)
        #     samples_nodes=torch.unique(torch.concat([samples_nodes,samples_nodes_i]))
        #     row=torch.concat([row,edge_index_i[0]])
        #     col=torch.concat([col,edge_index_i[1]])
        return samples_nodes, torch.stack([row, col], dim=0)

        # 不使用sample_from_node直接取所有点邻居方法：
        # row, col = edge_index
        # neighbors1=torch.concat([row[row==nodes[i]] for i in range(0, nodes.shape[0])])
        # neighbors2=torch.concat([col[row==nodes[i]] for i in range(0, nodes.shape[0])])
        # neighbors=torch.stack([neighbors1, neighbors2], dim=0)
        # print('neighbors: \n', neighbors)
